/*-----------------------------------------------------------------------

File  : ccl_splitting.c

Author: Stephan Schulz

Contents
 
  Implements functions for destructive splitting of clauses with at
  least two non-propositional variable disjoint subsets of literals.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Wed Apr 18 18:24:18 MET DST 2001
    New

-----------------------------------------------------------------------*/

#include "ccl_splitting.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: gen_prop_lit()
//
//   Generate a propositional literal with terms from bank.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static Eqn_p gen_prop_lit(TB_p bank, FunCode pred, bool positive)
{
   Term_p lside;
   Eqn_p  res;

   assert(bank);assert(pred > 0);
   assert(SigFindArity(bank->sig, pred) == 0);
   
   lside = TBTermtopInsert(bank, TermConstCellAlloc(pred));
   res = EqnAlloc(lside, bank->true_term, bank, positive);
   EqnSetProp(res, EPIsSplitLit);
   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: find_free_literal()
//
//   Find the first entry in lit_table that corresponds to a literal
//   not yet assigned to any clause part and return its index. If none
//   exists, return -1.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static int find_free_literal(LitSplitDesc_p lit_table, int lit_no)
{
   int res = -1,i;

   for(i=0; i<lit_no; i++)
   {
      if(!lit_table[i].part)
      {
	 res = i;
	 break;
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: build_part()
//
//   Given the index of the first unassigned literal in lit_table and
//   a part number, assign this number to all literals that are
//   transitively variable-linked to this first literal.
//
// Global Variables: -
//
// Side Effects    : Accumulates all part variables in
//                   lit_table[lit_index].varset 
//
/----------------------------------------------------------------------*/

static void build_part(LitSplitDesc_p lit_table, int lit_no, int
		       lit_index, int part)
{
   bool new_vars = true;
   int  i;

   lit_table[lit_index].part = part;
   while(new_vars)
   {
      new_vars = false;
      for(i=lit_index+1; i< lit_no; i++)
      {
	 if(!lit_table[i].part)
	 {
	    if(PTreeSharedElement(&(lit_table[lit_index].varset),
				  lit_table[i].varset))
	    {
	       lit_table[i].part = part;
	       new_vars = PTreeMerge(&(lit_table[lit_index].varset),
				     lit_table[i].varset)||new_vars;
	       lit_table[i].varset = NULL;
	    }
	 }
      }
   }
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: ClauseHasSplitLiteral()
//
//   Return true if a literal in the clause is a split literal.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseHasSplitLiteral(Clause_p clause)
{
   Eqn_p handle;

   for(handle = clause->literals; handle; handle = handle->next)
   {
      if(EqnQueryProp(handle, EPIsSplitLit))
      {
	 return true;
      }
   }
   return false;   
}




/*-----------------------------------------------------------------------
//
// Function: ClauseSplit()
//
//   Try to split clause into different clauses according to the
//   inference rule below. If successful,
//   deposit split clauses into set and return number of clauses
//   created. Otherwise return 0.
//
//   L1 v L2 v L3 ...
//   ----------------------------------------------------
//   T1 v T2 v T3 ...., ~T1 v L1, ~T2 v L2, ~T3 v L3, ...
//
//   if the Li are variable-disjoint subsets of the clause and the Ti
//   are _new_ propositional variables.
//
// Global Variables: -
//
// Side Effects    : Memory allocation, if successful, destroys
//                   clause!
//
/----------------------------------------------------------------------*/

int ClauseSplit(Clause_p clause, ClauseSet_p set, SplitType how, long*
		idents, float mod_weight)
{
   int            res = 0, part = 0,i,j,size, lit_no;
   LitSplitDesc_p lit_table;
   Eqn_p          handle,tmp, join;
   FunCode        new_pred;
   long           new_weight;
   TB_p           bank;
   Clause_p       new;

   assert(clause);
   assert(!clause->children);
   assert(!clause->set);
   assert(set);

   lit_no = ClauseLiteralNumber(clause);

   if(lit_no<=1 || ClauseHasSplitLiteral(clause))
   {
      return 0;
   }
   bank = clause->literals->bank;
   size = lit_no*sizeof(LitSplitDescCell);

   lit_table = SizeMalloc(size);

   handle = clause->literals;

   /* Initializing */
   for(i=0; i<lit_no; i++)
   {
      assert(handle);
      lit_table[i].literal = handle;
      lit_table[i].part    = 0;
      lit_table[i].varset  = NULL;
      j = EqnCollectVariables(handle, &(lit_table[i].varset)); 
                 /* j (mis)used as dummy to stiffle warnings - we are
		    only calling EqnCollectVariables for the side
		    effect here */
      if(how == SplitGroundOne || how == SplitGroundNone)
      {
	 if(!lit_table[i].varset)
	 {
	    lit_table[i].part = 1;
	 }
      }
      handle = handle->next;
   }
   if((how == SplitGroundOne) && find_free_literal(lit_table, lit_no))
   {
      part++;
   }
   /* Find the variable disjoint parts of the clause */
   while((i=find_free_literal(lit_table, lit_no))!=-1)
   {
      part++;
      build_part(lit_table, lit_no, i, part);
   }

   if(part>1)
   {

      /* fprintf(GlobalOut, "# SplitI: ");
      ClausePrint(GlobalOut, clause, true);
      fputc('\n', GlobalOut); */


      /* Build split clauses from original literals */
      join = NULL;
      clause->literals = NULL; /* Literals are recycled in new
				  clauses, clause skeleton is
				  deallocated below */
      for(i=1; i<=part; i++)
      {
	 new_pred = SigGetNewPredicateCode(bank->sig, 0);
	 new_weight = 0;
	 tmp    = gen_prop_lit(bank, new_pred, false);
	 tmp->next = join;
	 join = tmp;
	 handle = gen_prop_lit(bank, new_pred, true);
	 for(j=0; j<lit_no; j++)
	 {
	    if(lit_table[j].part == i)
	    {
	       tmp = lit_table[j].literal;
	       tmp->next = handle; 
	       handle = tmp;
	    }
	    else
	    {
	       new_weight += EqnSplitModStandardWeight(lit_table[j].literal);
	    }
	 }
	 new = ClauseAlloc(handle);
	 if(idents)
	 {
	    new->ident =  ++(*idents);
	 }
	 ClauseSetInsert(set, new);
	 /* if(OutputLevel>=1)
	 {
	    fprintf(GlobalOut, "# Split1: ");
	    ClausePrint(GlobalOut, new, true);
	    fputc('\n',GlobalOut);
	    }*/
	 ClauseDefaultDocumentSplitCreation(new, clause->ident);
	 SigSetSpecialWeight(bank->sig, new_pred, mod_weight*new_weight);
      }
      new = ClauseAlloc(join);
      if(idents)
      {
	 new->ident =  ++(*idents);
      }
      ClauseSetInsert(set, new);
      /* if(OutputLevel>=1)
      {
	 fprintf(GlobalOut, "# Split2: ");
	 ClausePrint(GlobalOut, new, true);
	 fputc('\n',GlobalOut);
	 }*/
      ClauseDefaultDocumentSplitCreation(new, clause->ident);
      ClauseDefaultDocumentStatusChange(clause, CS_DELETE);
      ClauseFree(clause);      /* We still retain the literals in the
				  split clauses! */
      res = part+1;
   }
   for(i=0; i<lit_no; i++)
   {
      PTreeFree(lit_table[i].varset);
   }
   SizeFree(lit_table, size);

   return res;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseSetSplitClauses()
//
//   Split all clauses in from_set and put the result into to_set.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long ClauseSetSplitClauses(ClauseSet_p from_set, ClauseSet_p to_set,
			  SplitType how, long* idents, float
			   mod_weight)
{
   long res=0, tmp;
   Clause_p handle;

   assert(from_set&&to_set);

   while(!ClauseSetEmpty(from_set))
   {
      handle = ClauseSetExtractFirst(from_set);
      tmp = ClauseSplit(handle, to_set, how, idents, mod_weight);
      if(!tmp)
      {
	 ClauseSetInsert(to_set, handle);
      }
      res += tmp;
   }
   return res;
}



/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


