/*-----------------------------------------------------------------------

File  : che_to_precgen.c

Author: Stephan Schulz

Contents
 
  Functions implementing several precedence generation schemes for
  term orderings.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Fri Sep 25 02:49:11 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "che_to_precgen.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

char* TOPrecGenNames[]=
{
   "none",             /* PNoMethod */
   "unary_first",      /* PUnaryFirst */
   "arity",            /* PArity */
   "invarity",         /* PInvArity */
   "const_max",        /* PConstFirst  */
   "const_min",         /* PInvArConstMin */
   "freq",             /* PByFrequency */
   "invfreq",          /* PByInvFrequency */
   "orient_axioms",    /* POrientAxioms */
   NULL
};


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/

typedef struct feature_sort_cell
{
   int     key1;
   int     key2;
   FunCode symbol;
}featuresortcell, *featuresort_p;

/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: feature_compare_function()
//
//   Compare two featuresortcells and return <0, =0, >0 as for
//   strcmp(). 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static int feature_compare_function(featuresort_p entry1,
				    featuresort_p entry2)
{
   int res;

   if((res = entry1->key1-entry2->key1))
   {
      return res;
   }
   if((res = entry1->key2-entry2->key2))
   {
      return res;
   }
   return entry1->symbol-entry2->symbol;
}


/*-----------------------------------------------------------------------
//
// Function: compute_precedence_from_array()
//
//   Given an ocb and a sorted array of type featuresortcell[], set
//   the precedence in the ocb.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void compute_precedence_from_array(OCB_p ocb, featuresort_p
					  array)
{
   FunCode i, last;
   
   last = SIG_TRUE_CODE;
   for(i = SIG_TRUE_CODE+1; i<= ocb->sig->f_count; i++)
   {
      OCBPrecedenceAddTuple(ocb, last, array[i].symbol, to_lesser);
      last = array[i].symbol;
   }
}



/*-----------------------------------------------------------------------
//
// Function: generate_unary_first_precedence()
//
//   Generate a precende in which symbols with higher arity are
//   larger, but unary symbols are larger still. Order of occurence in
//   the signature is used as a tie-breaker.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

static void generate_unary_first_precedence(OCB_p ocb)
{
   featuresort_p array =
      SizeMalloc((ocb->sig->f_count+1)*sizeof(featuresortcell));
   FunCode i;
   long    arity;

   for(i=SIG_BOTTOM_CODE; i<= ocb->sig->f_count; i++)
   {
      arity = SigFindArity(ocb->sig, i);
      if(arity == 1)
      {
	 array[i].key1 = INT_MAX;
      }
      else
      {
	 array[i].key1 = arity;
      }
      array[i].key2 = 0;
      array[i].symbol = i;
   }
   qsort(&array[SIG_TRUE_CODE+1], ocb->sig->f_count-SIG_TRUE_CODE,
	 sizeof(featuresortcell),
	 (ComparisonFunctionType)feature_compare_function);
   compute_precedence_from_array(ocb, array);
      
   SizeFree(array, (ocb->sig->f_count+1)*sizeof(featuresortcell));
}

/*-----------------------------------------------------------------------
//
// Function: generate_arity_precedence()
//
//   Generate a precende in which symbols with higher arity are
//   larger. Order of occurence in the signature is used as a
//   tie-breaker.  
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

static void generate_arity_precedence(OCB_p ocb)
{
   featuresort_p array =
      SizeMalloc((ocb->sig->f_count+1)*sizeof(featuresortcell));
   FunCode i;
   
   for(i=SIG_BOTTOM_CODE; i<= ocb->sig->f_count; i++)
   {
      array[i].key1 = SigFindArity(ocb->sig, i);
      array[i].key2 = 0;
      array[i].symbol = i;
   }
   qsort(&array[SIG_TRUE_CODE+1], ocb->sig->f_count-SIG_TRUE_CODE,
	 sizeof(featuresortcell),
	 (ComparisonFunctionType)feature_compare_function);
   compute_precedence_from_array(ocb, array);
      
   SizeFree(array, (ocb->sig->f_count+1)*sizeof(featuresortcell));
}


/*-----------------------------------------------------------------------
//
// Function: generate_invarity_precedence()
//
//   Generate a precende in which symbols with higher arity are
//   smaller. Order of occurence in the signature is used as a
//   tie-breaker.  
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

static void generate_invarity_precedence(OCB_p ocb)
{
   featuresort_p array =
      SizeMalloc((ocb->sig->f_count+1)*sizeof(featuresortcell));
   FunCode i;
   
   for(i=SIG_BOTTOM_CODE; i<= ocb->sig->f_count; i++)
   {
      array[i].key1 = -SigFindArity(ocb->sig, i);
      array[i].key2 = 0;
      array[i].symbol = i;
   }
   qsort(&array[SIG_TRUE_CODE+1], ocb->sig->f_count-SIG_TRUE_CODE,
	 sizeof(featuresortcell),
	 (ComparisonFunctionType)feature_compare_function);
   compute_precedence_from_array(ocb, array);
      
   SizeFree(array, (ocb->sig->f_count+1)*sizeof(featuresortcell));
}


/*-----------------------------------------------------------------------
//
// Function: generate_const_max_precedence()
//
//   Generate a precedence in which symbols with higher arity are
//   larger, but constants are the largest symbols. Order of occurence
//   in the signature is used as a tie-breaker.  
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

static void generate_const_max_precedence(OCB_p ocb)
{
   featuresort_p array =
      SizeMalloc((ocb->sig->f_count+1)*sizeof(featuresortcell));
   FunCode i;

   for(i=SIG_BOTTOM_CODE; i<= ocb->sig->f_count; i++)
   {
      array[i].key1 = SigFindArity(ocb->sig, i);
      if(!array[i].key1)
      {
	 array[i].key1 = INT_MAX;
      }
      array[i].key2 = 0;
      array[i].symbol = i;
   }
   qsort(&array[SIG_TRUE_CODE+1], ocb->sig->f_count-SIG_TRUE_CODE,
	 sizeof(featuresortcell),
	 (ComparisonFunctionType)feature_compare_function);
   compute_precedence_from_array(ocb, array);
      
   SizeFree(array, (ocb->sig->f_count+1)*sizeof(featuresortcell));
}


/*-----------------------------------------------------------------------
//
// Function: generate_const_min_precedence()
//
//   Generate a precedence in which symbols with higher arity are
//   smaller, but constants are the smallest symbols. Order of
//   occurence in the signature is used as a tie-breaker.  
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

static void generate_const_min_precedence(OCB_p ocb)
{
   featuresort_p array =
      SizeMalloc((ocb->sig->f_count+1)*sizeof(featuresortcell));
   FunCode i;

   for(i=SIG_BOTTOM_CODE; i<= ocb->sig->f_count; i++)
   {
      array[i].key1 = -SigFindArity(ocb->sig, i);
      if(!array[i].key1)
      {
	 array[i].key1 = INT_MIN;
      }
      array[i].key2 = 0;
      array[i].symbol = i;
   }
   qsort(&array[SIG_TRUE_CODE+1], ocb->sig->f_count-SIG_TRUE_CODE,
	 sizeof(featuresortcell),
	 (ComparisonFunctionType)feature_compare_function);
   compute_precedence_from_array(ocb, array);
      
   SizeFree(array, (ocb->sig->f_count+1)*sizeof(featuresortcell));
}



/*-----------------------------------------------------------------------
//
// Function: generate_freq_precedence()
//
//   Generate a precedence in which symbols which occur more often in
//   the specification are bigger. Arity is used as a tie-breaker,
//   then order of occurence in the signature. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

static void generate_freq_precedence(OCB_p ocb, ClauseSet_p axioms)
{
   featuresort_p array;
   FunCode       i;
   long          *dist_array;

   array = SizeMalloc((ocb->sig->f_count+1)*sizeof(featuresortcell));
   dist_array = SizeMalloc((ocb->sig->f_count+1) * sizeof(long));

   for(i=0; i<=ocb->sig->f_count; i++)
   {
      dist_array[i] = 0;
   }      
   ClauseSetAddSymbolDistribution(axioms, dist_array);
   
   for(i=SIG_BOTTOM_CODE; i<= ocb->sig->f_count; i++)
   {
      array[i].key1 = dist_array[i];
      array[i].key2 = SigFindArity(ocb->sig, i);
      array[i].symbol = i;
   }
   qsort(&array[SIG_TRUE_CODE+1], ocb->sig->f_count-SIG_TRUE_CODE,
	 sizeof(featuresortcell),
	 (ComparisonFunctionType)feature_compare_function);
   compute_precedence_from_array(ocb, array);
      
   SizeFree(dist_array, (ocb->sig->f_count+1)*sizeof(long));      
   SizeFree(array, (ocb->sig->f_count+1)*sizeof(featuresortcell));
}



/*-----------------------------------------------------------------------
//
// Function: generate_invfreq_precedence()
//
//   Generate a precedence in which symbols which occur more often in
//   the specification are smaller. Arity is used as a tie-breaker,
//   then order of occurence in the signature. .  
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

static void generate_invfreq_precedence(OCB_p ocb, ClauseSet_p axioms)
{
   featuresort_p array;
   FunCode       i;
   long          *dist_array;

   array = SizeMalloc((ocb->sig->f_count+1)*sizeof(featuresortcell));
   dist_array = SizeMalloc((ocb->sig->f_count+1) * sizeof(long));

   for(i=0; i<=ocb->sig->f_count; i++)
   {
      dist_array[i] = 0;
   }      
   ClauseSetAddSymbolDistribution(axioms, dist_array);
   
   for(i=SIG_BOTTOM_CODE; i<= ocb->sig->f_count; i++)
   {
      array[i].key1 = -dist_array[i];
      array[i].key2 = SigFindArity(ocb->sig, i);
      array[i].symbol = i;
   }
   qsort(&array[SIG_TRUE_CODE+1], ocb->sig->f_count-SIG_TRUE_CODE,
	 sizeof(featuresortcell),
	 (ComparisonFunctionType)feature_compare_function);
   compute_precedence_from_array(ocb, array);
      
   SizeFree(dist_array, (ocb->sig->f_count+1)*sizeof(long));      
   SizeFree(array, (ocb->sig->f_count+1)*sizeof(featuresortcell));
}





/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/



/*-----------------------------------------------------------------------
//
// Function: TOTranslatePrecGenMethod()
//
//   Given a string, return the corresponding TOPrecGenMethod
//   token. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

TOPrecGenMethod TOTranslatePrecGenMethod(char* name)
{
   TOPrecGenMethod method;
   
   method = StringIndex(name, TOPrecGenNames);

   if(method == -1)
   {
      method = PNoMethod;
   }
   return method;
}


/*-----------------------------------------------------------------------
//
// Function: TOGeneratePrecedence()
//
//   Given a pre-initialized OCB, compute a good precedence for a term
//   ordering.
//
// Global Variables: -
//
// Side Effects    : Sets weights in ocb.
//
/----------------------------------------------------------------------*/

void TOGeneratePrecedence(OCB_p ocb, ClauseSet_p axioms,
			  char* predefined, TOPrecGenMethod method)
{  
   assert(ocb);
   assert(ocb->precedence);
   assert(ocb->sig);

   OCBPrecedenceAddTuple(ocb, SIG_TRUE_CODE, SIG_BOTTOM_CODE, to_greater);
   
   if(predefined)
   {
      Scanner_p in = CreateScanner(StreamTypeUserString, predefined,
				   true, NULL);

      TOPrecedenceParse(in, ocb);
      
      DestroyScanner(in);
   }
   
   switch(method)
   {
   case POrientAxioms:
	 Error("Not yet implemented", OTHER_ERROR);
	 break;
   case PNoMethod:	 
	 if(predefined) /* User should know what he does now! */
	 {
	    break;
	 } /* else: Fall through */
   case PUnaryFirst:
	 generate_unary_first_precedence(ocb);
	 break;
   case PArity:
	 generate_arity_precedence(ocb);
	 break;
   case PInvArity:
	 generate_invarity_precedence(ocb);
	 break;
   case PConstMax:
	 generate_const_max_precedence(ocb);
	 break;
   case PInvArConstMin:
	 generate_const_min_precedence(ocb);
	 break;
   case PByFrequency:
	 generate_freq_precedence(ocb, axioms);
	 break;
   case PByInvFrequency:
	 generate_invfreq_precedence(ocb, axioms);
	 break;
   default:
	 assert(false && "Precedence generation method unimplemented");
	 break;
   }  
   assert(OCBFunCompare(ocb, SIG_BOTTOM_CODE, SIG_TRUE_CODE) ==
	  to_lesser);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


